﻿/*
 * 24 Puzzle
 * Copyright (C) 2008 Michael Birken
 * 
 * This file is part of 24 Puzzle.
 *
 * 24 Puzzle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * 24 Puzzle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TwentyFourPuzzle {
  public class Example5 {

    private void Solve(
        Number[] numbers,
        UnaryOperator[] unaryOperators,
        BinaryOperator[] binaryOperators,
        List<IElement> elements,
        bool[] numbersAvailable,
        bool[] unaryOperatorsAvailable,
        int stackSize,
        Stack<double> numberStack,
        Stack<string> stringStack,
        SortedDictionary<int, List<IElement>> expressions) {

      bool remainingNumbers = false;
      double lastNumber = Double.NaN;
      for (int i = 0; i < numbersAvailable.Length; i++) {
        if (numbersAvailable[i] && numbers[i].value != lastNumber) {
          remainingNumbers = true;
          elements.Add(numbers[i]);
          numbersAvailable[i] = false;
          Solve(numbers, unaryOperators, binaryOperators, elements,
              numbersAvailable, unaryOperatorsAvailable, stackSize + 1,
              numberStack, stringStack, expressions);
          numbersAvailable[i] = true;
          elements.RemoveAt(elements.Count - 1);
        }
      }

      if (stackSize >= 1) {
        for (int i = 0; i < unaryOperatorsAvailable.Length; i++) {
          if (unaryOperatorsAvailable[i]) {
            elements.Add(unaryOperators[i]);
            unaryOperatorsAvailable[i] = false;
            Solve(numbers, unaryOperators, binaryOperators, elements,
                numbersAvailable, unaryOperatorsAvailable, stackSize,
                numberStack, stringStack, expressions);
            unaryOperatorsAvailable[i] = true;
            elements.RemoveAt(elements.Count - 1);
          }
        }
      }

      if (stackSize >= 2) {
        for (int i = 0; i < binaryOperators.Length; i++) {
          elements.Add(binaryOperators[i]);
          Solve(numbers, unaryOperators, binaryOperators, elements,
              numbersAvailable, unaryOperatorsAvailable, stackSize - 1,
              numberStack, stringStack, expressions);
          elements.RemoveAt(elements.Count - 1);
        }
      }

      if (stackSize == 1 && !remainingNumbers) {
        double value = EvalutateExpression(elements, numberStack);
        int floored = (int)Math.Floor(value);
        if (value == floored) {
          if (expressions.ContainsKey(floored)) {
            if (elements.Count < expressions[floored].Count) {
              expressions[floored] = new List<IElement>(elements);
            }
          } else {
            expressions[floored] = new List<IElement>(elements);
          }
        }
      }
    }

    private double EvalutateExpression(List<IElement> elements,
        Stack<double> stack) {
      foreach (IElement element in elements) {
        element.Evaluate(stack);
      }
      return stack.Pop();
    }

    private void PrintExpression(List<IElement> elements, double target,
        Stack<string> stack) {
      foreach (IElement element in elements) {
        element.Print(stack);
      }
      string expression = stack.Pop();
      Console.WriteLine("{0} = {1}", target, expression);
    }

    public void FindAll(double[] numbers, UnaryOperator[] unaryOperators,
        BinaryOperator[] binaryOperators) {
      Array.Sort(numbers);
      Number[] nums = new Number[numbers.Length];
      for (int i = 0; i < numbers.Length; i++) {
        nums[i] = new Number(numbers[i]);
      }
      List<IElement> elements = new List<IElement>();
      bool[] numbersAvailable = new bool[numbers.Length];
      for (int i = 0; i < numbersAvailable.Length; i++) {
        numbersAvailable[i] = true;
      }
      bool[] unaryOperatorsAvailable = new bool[unaryOperators.Length];
      for (int i = 0; i < unaryOperatorsAvailable.Length; i++) {
        unaryOperatorsAvailable[i] = true;
      }
      Stack<double> numberStack = new Stack<double>();
      Stack<string> stringStack = new Stack<string>();
      SortedDictionary<int, List<IElement>> expressions 
          = new SortedDictionary<int, List<IElement>>();

      Solve(nums, unaryOperators, binaryOperators, elements,
          numbersAvailable, unaryOperatorsAvailable, 0, numberStack,
          stringStack, expressions);

      foreach (KeyValuePair<int, List<IElement>> pair in expressions) {
        PrintExpression(pair.Value, pair.Key, stringStack);
      }
    }
  }
}
